@@ -4,15 +4,15 @@
 #include "EXTERN.h"
 #include "perl.h"
 #include "assert.h"
+#include "srl_reader_types.h"
 
 typedef struct PTABLE * ptable_ptr;
-typedef struct {
-    const unsigned char *buf_start;           /* ptr to "physical" start of input buffer */
-    const unsigned char *buf_end;             /* ptr to end of input buffer */
-    const unsigned char *pos;                 /* ptr to current position within input buffer */
-    const unsigned char *save_pos;            /* used for COPY tags */
-    const unsigned char *body_pos;            /* in Sereal V2, all offsets are relative to the body */
-    STRLEN buf_len;
+typedef struct srl_decoder srl_decoder_t;
+
+struct srl_decoder {
+    srl_reader_buffer_t buf;
+    srl_reader_buffer_ptr pbuf;
+    const unsigned char *save_pos;      /* used for COPY tags */
 
     U32 flags;                          /* flag-like options: See SRL_F_DECODER_* defines in srl_decoder.c */
     UV max_recursion_depth;             /* Configurable limit on the number of recursive calls we're willing to make */
@@ -31,7 +31,7 @@ typedef struct {
     U8 proto_version;
     U8 encoding_flags;
     U32 flags_readonly;
-} srl_decoder_t;
+};
 
 typedef struct {
     SV *sv;
@@ -51,6 +51,8 @@ SV *srl_decode_into(pTHX_ srl_decoder_t *dec, SV *src, SV *body_into, UV start_o
 SV *srl_decode_header_into(pTHX_ srl_decoder_t *dec, SV *src, SV *header_into, UV start_offset);
 /* decode both header and body - must pass in two SVs to write into */
 void srl_decode_all_into(pTHX_ srl_decoder_t *dec, SV *src, SV *header_into, SV *body_into, UV start_offset);
+/* main recursive dump routine, for internal usage only!!! */
+void srl_decode_single_value(pTHX_ srl_decoder_t *dec, SV* into, SV** container);
 
 /* Explicit destructor */
 void srl_destroy_decoder(pTHX_ srl_decoder_t *dec);
@@ -58,26 +60,6 @@ void srl_destroy_decoder(pTHX_ srl_decoder_t *dec);
 /* destructor hook - called automagically */
 void srl_decoder_destructor_hook(pTHX_ void *p);
 
-#define BUF_POS(dec) ((dec)->pos)
-#define BUF_SPACE(dec) ((dec)->buf_end - (dec)->pos)
-#define BUF_POS_OFS(dec) ((dec)->pos - (dec)->buf_start)
-#define BUF_SIZE(dec) ((dec)->buf_end - (dec)->buf_start)
-#define BUF_NOT_DONE(dec) ((dec)->pos < (dec)->buf_end)
-#define BUF_DONE(dec) ((dec)->pos >= (dec)->buf_end)
-
-#define BODY_POS_OFS(enc) ((dec)->pos - (dec)->body_pos)
-
-/* these are mostly for right between deserializing the header and the body */
-#define SRL_SET_BODY_POS(dec, pos_ptr) ((dec)->body_pos = pos_ptr)
-#define SRL_UPDATE_BODY_POS(dec)                                                    \
-    STMT_START {                                                                    \
-        if (expect_false(SRL_DEC_HAVE_OPTION((dec), SRL_F_DECODER_PROTOCOL_V1))) {  \
-            SRL_SET_BODY_POS(dec, (dec)->buf_start);                                \
-        } else {                                                                    \
-            SRL_SET_BODY_POS(dec, (dec)->pos-1);                                    \
-        }                                                                           \
-    } STMT_END
-
 /* Macro to assert that the type of an SV is complex enough to
  * be an RV. Differs on old perls since there used to be an RV type.
  */
@@ -101,14 +83,6 @@ void srl_decoder_destructor_hook(pTHX_ void *p);
         }                                                           \
     } STMT_END
 
-#define SRL_sv_set_rv_to(into,referent)             \
-    STMT_START {                                    \
-        SRL_prepare_SV_for_RV(into);                \
-        SvTEMP_off(referent);                       \
-        SvRV_set(into, referent);                   \
-        SvROK_on(into);                             \
-    } STMT_END
-
 /* If set, the decoder struct needs to be cleared instead of freed at
  * the end of a deserialization operation */
 #define SRL_F_REUSE_DECODER                     0x00000001UL
